/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <string.h>

#include <glib.h>

#include <pan/base/debug.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pan-i18n.h>

#include <pan/filters/filter-phrase.h>

const gchar * FILTER_PHRASE_CLASS_ID = "PanObject::Filter::FilterPhrase";

/************
*************  
************/

#define is_metacharacter(A) (_metacharacters[(guchar)(A)])
#if 0
static char _metacharacters[UCHAR_MAX];
#define PRINT_TABLE(A) \
	printf ("static char " #A "[UCHAR_MAX] = {"); \
	for (i=0; i<UCHAR_MAX; ++i) { \
		if (!(i%40)) \
			printf ("\n\t"); \
		printf ("%d,", A[i]); \
	} \
	printf ("};\n\n");
static void
build_table (void)
{
	int i;
	unsigned char ch;

	for (ch=0; ch<UCHAR_MAX; ++ch) {
		_metacharacters[ch] = ch=='.' || ch=='^' || ch=='$' || ch=='*' ||
		                      ch=='+' || ch=='?' || ch=='{' || ch=='[' ||
				      ch=='|' || ch=='(' || ch==')' || ch=='\\';
	}

	PRINT_TABLE(_metacharacters)
}
#else
static char _metacharacters[UCHAR_MAX] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,
	1,1,1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,};
#endif


/************
*************  PROTECTED
************/

static Filter *
filter_phrase_dup (const Filter * f_old)
{
	Filter * f_new = filter_phrase_new ();
	FilterPhrase * fp_old = FILTER_PHRASE(f_old);
	FilterPhrase * fp_new = FILTER_PHRASE(f_new);
	filter_class_dup (f_old, f_new);
	filter_phrase_set (fp_new, fp_old->public_match_type, fp_old->key_type, fp_old->public_key, fp_old->case_sensitive);
	return f_new;
}

static char*
filter_phrase_to_string (const Filter * filter)
{
	const char * key_str = "";
	const char * match_str = "";
	const gboolean negate = filter->negate;
	const FilterPhrase * phrase = FILTER_PHRASE(filter);

	switch (phrase->key_type) {
		case PHRASE_KEY_SUBJECT:     key_str = _("Subject"); break;
		case PHRASE_KEY_AUTHOR:      key_str = _("Author"); break;
		case PHRASE_KEY_MESSAGE_ID:  key_str = _("Message-ID"); break;
		case PHRASE_KEY_REFERENCES:  key_str = _("References"); break;
		case PHRASE_KEY_XREF:        key_str = _("Xref"); break;
		default:                     g_warning ("Unrecognized key_type %d", phrase->key_type);
	}

	if (!negate) {
		switch (phrase->public_match_type) {
			case PHRASE_MATCH_CONTAINS:     match_str = _("contains"); break;
			case PHRASE_MATCH_IS:           match_str = _("is"); break;
			case PHRASE_MATCH_STARTS_WITH:  match_str = _("starts with"); break;
			case PHRASE_MATCH_ENDS_WITH:    match_str = _("ends with"); break;
			case PHRASE_MATCH_REGEX:        match_str = _("matches regular expression"); break;
		        default:                        g_warning ("Unrecognized match_type %d", phrase->public_match_type);
		}
	} else {
		switch (phrase->public_match_type) {
			case PHRASE_MATCH_CONTAINS:     match_str = _("does not contain"); break;
			case PHRASE_MATCH_IS:           match_str = _("is not"); break;
			case PHRASE_MATCH_STARTS_WITH:  match_str = _("does not start with"); break;
			case PHRASE_MATCH_ENDS_WITH:    match_str = _("does not end with"); break;
			case PHRASE_MATCH_REGEX:        match_str = _("does not match regular expression"); break;
		        default:                        g_warning ("Unrecognized match_type %d", phrase->public_match_type);
		}
	}

	return g_strdup_printf ("%s %s \"%s\"", key_str, match_str, phrase->public_key);
}

static int
my_regexec (FilterPhrase * filter, const char * compare_str)
{
	if (!filter->regex_compiled)
	{
		int flags = REG_EXTENDED;
		if (!filter->case_sensitive)
			flags |= REG_ICASE;
		regcomp (&filter->key_regex, filter->private_key, flags);

		filter->regex_compiled = TRUE;
	}

	return regexec (&filter->key_regex, compare_str, 0, NULL, 0);
}

static void
filter_phrase_test_articles (Filter          * filter,
                             const Article  ** articles,
                             int               article_qty,
                             gboolean        * does_match)
{
	int i;
	FilterPhrase * phrase = FILTER_PHRASE(filter);
	GString * compare = g_string_new (NULL);

	for (i=0; i<article_qty; ++i)
	{
		char buf[512];
		const Article * a = articles[i];
		const char * compare_str = "";

		switch (phrase->key_type)
		{
			case PHRASE_KEY_SUBJECT:     compare_str = article_get_subject (a); break;
			case PHRASE_KEY_AUTHOR:      compare_str = article_get_author_str (a, buf, sizeof(buf)); break;
			case PHRASE_KEY_MESSAGE_ID:  compare_str = article_get_message_id (a); break;
			case PHRASE_KEY_REFERENCES:  compare_str = a->references; break;
			case PHRASE_KEY_XREF:        compare_str = a->xref; break;
		}

		if (compare_str == NULL)
		{
			does_match[i] = FALSE;
		}
		else if (phrase->match_type == PHRASE_MATCH_REGEX)
		{
			does_match[i] = !my_regexec (phrase, compare_str);
		}
		else
		{
			const char * key = NULL;

			key = phrase->case_sensitive
				? phrase->private_key
				: phrase->private_key_lower;

			if (!compare_str)
				g_string_truncate (compare, 0);
			else {
				g_string_assign (compare, compare_str);
			       	if (!phrase->case_sensitive)
					g_string_ascii_down (compare);
			}

			if (phrase->match_type == PHRASE_MATCH_IS)
			{
				does_match[i] = !pan_strcmp (compare->str, key);
			}
			else
			{
				const char * pch = strstr (compare->str, key);

				if (phrase->match_type == PHRASE_MATCH_CONTAINS)
				{
					does_match[i] = pch != NULL;
				}
				else if (phrase->match_type == PHRASE_MATCH_STARTS_WITH)
				{
					does_match[i] = pch == compare->str;
				}
				else if (phrase->match_type == PHRASE_MATCH_ENDS_WITH)
				{
					does_match[i] = pch == compare->str + compare->len - phrase->private_key_len;
				}
			}
		}
	}
	
	g_string_free (compare, TRUE);
}

static void
filter_phrase_destructor (PanObject * o)
{
	filter_phrase_set (FILTER_PHRASE(o), PHRASE_MATCH_IS, PHRASE_KEY_SUBJECT, NULL, FALSE);
	filter_destructor (o);
}

static void
filter_phrase_constructor (FilterPhrase * f)
{
	debug_enter ("filter_phase_constructor");

	filter_constructor ((Filter*)f,
	                    filter_phrase_destructor,
	                    filter_phrase_test_articles,
	                    filter_phrase_to_string,
	                    filter_phrase_dup,
	                    FILTER_PHRASE_CLASS_ID);

	f->public_match_type = PHRASE_MATCH_IS;
	f->match_type = PHRASE_MATCH_IS;
	f->key_type = PHRASE_KEY_SUBJECT;
	f->private_key_lower = NULL;
	f->private_key_len = 0;
	f->private_key = NULL;
	f->public_key = NULL;
	f->case_sensitive = FALSE;
	f->regex_compiled = FALSE;

	debug_exit ("filter_phase_constructor");
}

static char*
regexp_unescape (const char * in)
{
	char * retval = g_new (char, strlen(in)+1);
	char * out = retval;
	if (*in == '^')
		++in;
	while (*in) {
		if (in[0]=='\\' && is_metacharacter(in[1]))
			++in;
		*out++ = *in++;
	}
	if (out!=retval && out[-1]=='$')
		--out;
	*out = '\0';
	return retval;
}

/**
 * Try to downgrade the more-expensive regexes to a cheaper type.
 */
static PhraseMatchType
get_real_match_type (const char * key, PhraseMatchType type)
{
	gboolean starts_with = FALSE;
	gboolean ends_with = FALSE;
	const char * pch;
	const char * end;
	size_t len;

	/* sanity clause */
	g_return_val_if_fail (is_nonempty_string(key), type);

	/* if it's not a regex, keep it */
	if (type != PHRASE_MATCH_REGEX)
		return type;

	/* must it be a regex? */
	len = strlen (key);
	end = key + len;
	for (pch=key; pch!=end; ++pch)
		if (*pch=='\\' && is_metacharacter(pch[1]))
			++pch;
		else if (*pch=='^' && pch==key)
			starts_with = TRUE;
		else if (*pch=='$' && pch+1==end)
			ends_with = TRUE;
		else if (is_metacharacter(*pch))
			return PHRASE_MATCH_REGEX;

	if (starts_with && ends_with)
		return PHRASE_MATCH_IS;
	if (starts_with)
		return PHRASE_MATCH_STARTS_WITH;
	if (ends_with)
		return PHRASE_MATCH_ENDS_WITH;
	return PHRASE_MATCH_CONTAINS;
}

/************
*************  PUBLIC
************/

Filter*
filter_phrase_new (void)
{
	FilterPhrase * f;
	debug_enter ("filter_phrase_new");

	f = g_new0 (FilterPhrase, 1);
	filter_phrase_constructor (f);

	debug_exit ("filter_phrase_new");
	return FILTER(f);
}

void
filter_phrase_set (FilterPhrase        * filter,
                   PhraseMatchType       match_type,
                   PhraseKeyType         key_type,
                   const char          * key,
                   gboolean              case_sensitive)
{
	debug_enter ("filter_phrase_set");

	/* sanity clause */
	g_return_if_fail (filter!=NULL);

	/* free the previous phrase */
	replace_gstr (&filter->public_key, NULL);
	replace_gstr (&filter->private_key, NULL);
	replace_gstr (&filter->private_key_lower, NULL);
	if (filter->regex_compiled) {
		regfree (&filter->key_regex);
		filter->regex_compiled = FALSE;
	}

	/* repopulate the filter */
	filter->public_match_type = match_type;
	filter->match_type = key==NULL ? match_type : get_real_match_type (key, match_type);
	filter->key_type = key_type;
	filter->case_sensitive = case_sensitive;
	filter->private_key_len = 0;
	if (key != NULL)
	{
		/* build the key strings */
		filter->public_key = g_strdup (key);
		filter->private_key = filter->match_type == filter->public_match_type
			? g_strdup (key)
			: regexp_unescape (key);
		filter->private_key_lower = g_utf8_strdown (filter->private_key, -1);
		filter->private_key_len = strlen (filter->private_key);
	}

	debug_exit ("filter_phrase_set");
}
